@@ -31,6 +31,7 @@ SEARCH_SUGGEST_URL = FDFE + "searchSuggest"
 BULK_URL = FDFE + "bulkDetails"
 LOG_URL = FDFE + "log"
 TOC_URL = FDFE + "toc"
+ACCEPT_TOS_URL = FDFE + "acceptTos"
 LIST_URL = FDFE + "list"
 REVIEWS_URL = FDFE + "rev"
 
@@ -54,7 +55,11 @@ class RequestError(Exception):
         return repr(self.value)
 
 class SecurityCheckError(Exception):
-    pass
+    def __init__(self, value):
+        self.value = value
+
+    def __str__(self):
+        return repr(self.value)
 
 
 class GooglePlayAPI(object):
@@ -63,23 +68,25 @@ class GooglePlayAPI(object):
     Usual APIs methods are login(), search(), details(), bulkDetails(),
     download(), browse(), reviews() and list()."""
 
-    def __init__(self, locale, timezone, device_codename='bacon',
+    def __init__(self, locale="en_US", timezone="UTC", device_codename="bacon",
                  proxies_config=None):
         self.authSubToken = None
         self.gsfId = None
         self.device_config_token = None
+        self.deviceCheckinConsistencyToken = None
+        self.dfeCookie = None
         self.proxies_config = proxies_config
         self.deviceBuilder = config.DeviceBuilder(device_codename)
-        self.set_locale(locale)
-        self.set_timezone(timezone)
+        self.setLocale(locale)
+        self.setTimezone(timezone)
 
-    def set_locale(self, locale):
-        self.deviceBuilder.set_locale(locale)
+    def setLocale(self, locale):
+        self.deviceBuilder.setLocale(locale)
 
-    def set_timezone(self, timezone):
-        self.deviceBuilder.set_timezone(timezone)
+    def setTimezone(self, timezone):
+        self.deviceBuilder.setTimezone(timezone)
 
-    def encrypt_password(self, login, passwd):
+    def encryptPassword(self, login, passwd):
         """Encrypt credentials using the google publickey, with the
         RSA algorithm"""
 
@@ -137,6 +144,10 @@ class GooglePlayAPI(object):
             headers["Authorization"] = "GoogleLogin auth=%s" % self.authSubToken
         if self.device_config_token is not None:
             headers["X-DFE-Device-Config-Token"] = self.device_config_token
+        if self.deviceCheckinConsistencyToken is not None:
+            headers["X-DFE-Device-Checkin-Consistency-Token"] = self.deviceCheckinConsistencyToken
+        if self.dfeCookie is not None:
+            headers["X-DFE-Cookie"] = self.dfeCookie
         return headers
 
     def checkin(self, email, ac2dmToken):
@@ -151,6 +162,7 @@ class GooglePlayAPI(object):
                             proxies=self.proxies_config)
         response = googleplay_pb2.AndroidCheckinResponse()
         response.ParseFromString(res.content)
+        self.deviceCheckinConsistencyToken = response.deviceCheckinConsistencyToken
 
         # checkin again to upload gfsid
         request.id = response.androidId
@@ -199,7 +211,7 @@ class GooglePlayAPI(object):
             # First time setup, where we obtain an ac2dm token and
             # upload device information
 
-            encryptedPass = self.encrypt_password(email, password).decode('utf-8')
+            encryptedPass = self.encryptPassword(email, password).decode('utf-8')
             # AC2DM token
             params = self.deviceBuilder.getLoginParams(email, encryptedPass)
             params['service'] = 'ac2dm'
@@ -220,7 +232,9 @@ class GooglePlayAPI(object):
                 ac2dmToken = params["auth"]
             elif "error" in params:
                 if "NeedsBrowser" in params["error"]:
-                    raise SecurityCheckError()
+                    raise SecurityCheckError("Security check is needed, try to visit "
+                                     "https://accounts.google.com/b/0/DisplayUnlockCaptcha "
+                                     "to unlock, or setup an app-specific password")
                 raise LoginError("server says: " + params["error"])
             else:
                 raise LoginError("Auth token not found.")
@@ -233,7 +247,7 @@ class GooglePlayAPI(object):
             self.gsfId = gsfId
             self.setAuthSubToken(authSubToken)
             # check if token is valid with a simple search
-            self.search('firefox', 1, None)
+            self.search('drv')
         else:
             raise LoginError('Either (email,pass) or (gsfId, authSubToken) is needed')
 
@@ -297,7 +311,7 @@ class GooglePlayAPI(object):
 
     def executeRequestApi2(self, path, post_data=None, content_type=CONTENT_TYPE_URLENC, params=None):
         if self.authSubToken is None:
-            raise Exception("You need to login before executing any request")
+            raise LoginError("You need to login before executing any request")
         headers = self.getHeaders()
         headers["Content-Type"] = content_type
 
@@ -329,62 +343,29 @@ class GooglePlayAPI(object):
                   "ssis": "120",
                   "sst": "2"}
         data = self.executeRequestApi2(SEARCH_SUGGEST_URL, params=params)
-        response = data.payload.searchSuggestResponse
-        return [{"type": e.type,
-                 "suggestedQuery": e.suggestedQuery,
-                 "title": e.title} for e in response.entry]
+        entryIterator = data.payload.searchSuggestResponse.entry
+        return list(map(utils.parseProtobufObj, entryIterator))
 
-    def search(self, query, nb_result, offset=None):
+    def search(self, query):
         """ Search the play store for an app.
 
-        nb_result is the maximum number of result to be returned.
+        nb_result (int): is the maximum number of result to be returned
 
-        offset is used to take result starting from an index.
+        offset (int): is used to take result starting from an index.
         """
         if self.authSubToken is None:
-            raise Exception("You need to login before executing any request")
+            raise LoginError("You need to login before executing any request")
 
-        remaining = nb_result
-        output = []
-
-        nextPath = SEARCH_URL + "?c=3&q={}".format(requests.utils.quote(query))
-        if (offset is not None):
-            nextPath += "&o={}".format(offset)
-        while remaining > 0 and nextPath is not None:
-            currentPath = nextPath
-            data = self.executeRequestApi2(currentPath)
-            if utils.hasPrefetch(data):
-                response = data.preFetch[0].response
-            else:
-                response = data
-            if utils.hasSearchResponse(response.payload):
-                # we still need to fetch the first page, so go to
-                # next loop iteration without decrementing counter
-                nextPath = FDFE + response.payload.searchResponse.nextPageUrl
-                continue
-            if utils.hasListResponse(response.payload):
-                cluster = response.payload.listResponse.cluster
-                if len(cluster) == 0:
-                    # unexpected behaviour, probably due to expired token
-                    raise LoginError('Unexpected behaviour, probably expired '
-                                     'token')
-                cluster = cluster[0]
-                if len(cluster.doc) == 0:
-                    break
-                if cluster.doc[0].containerMetadata.nextPageUrl != "":
-                    nextPath = FDFE + cluster.doc[0].containerMetadata.nextPageUrl
-                else:
-                    nextPath = None
-                apps = []
-                for doc in cluster.doc:
-                    apps.extend(doc.child)
-                output += list(map(utils.fromDocToDictionary, apps))
-                remaining -= len(apps)
-
-        if len(output) > nb_result:
-            output = output[:nb_result]
-
-        return output
+        path = SEARCH_URL + "?c=3&q={}".format(requests.utils.quote(query))
+        # FIXME: not sure if this toc call should be here
+        self.toc()
+        data = self.executeRequestApi2(path)
+        if utils.hasPrefetch(data):
+            response = data.preFetch[0].response
+        else:
+            response = data
+        resIterator = response.payload.listResponse.doc
+        return list(map(utils.parseProtobufObj, resIterator))
 
     def details(self, packageName):
         """Get app details from a package name.
@@ -392,7 +373,7 @@ class GooglePlayAPI(object):
         packageName is the app unique ID (usually starting with 'com.')."""
         path = DETAILS_URL + "?doc={}".format(requests.utils.quote(packageName))
         data = self.executeRequestApi2(path)
-        return utils.fromDocToDictionary(data.payload.detailsResponse.docV2)
+        return utils.parseProtobufObj(data.payload.detailsResponse.docV2)
 
     def bulkDetails(self, packageNames):
         """Get several apps details from a list of package names.
@@ -418,19 +399,20 @@ class GooglePlayAPI(object):
                                           params=params)
         response = message.payload.bulkDetailsResponse
         return [None if not utils.hasDoc(entry) else
-                utils.fromDocToDictionary(entry.doc)
+                utils.parseProtobufObj(entry.doc)
                 for entry in response.entry]
 
-    def getHomeApps(self):
+    def home(self, cat=None):
         path = HOME_URL + "?c=3&nocache_isui=true"
+        if cat is not None:
+            path += "&cat={}".format(cat)
         data = self.executeRequestApi2(path)
-        output = []
-        cluster = data.preFetch[0].response.payload.listResponse.cluster[0]
-        for doc in cluster.doc:
-            output.append({"categoryId": doc.docid,
-                           "categoryStr": doc.title,
-                           "apps": [utils.fromDocToDictionary(c) for c in doc.child]})
-        return output
+        if utils.hasPrefetch(data):
+            response = data.preFetch[0].response
+        else:
+            response = data
+        resIterator = response.payload.listResponse.doc
+        return list(map(utils.parseProtobufObj, resIterator))
 
     def browse(self, cat=None, subCat=None):
         """Browse categories. If neither cat nor subcat are specified,
@@ -443,65 +425,46 @@ class GooglePlayAPI(object):
             path += "&ctr={}".format(requests.utils.quote(subCat))
         data = self.executeRequestApi2(path)
 
-        if cat is None and subCat is None:
-            # result contains all categories available
-            return [{'name': c.name,
-                     'dataUrl': c.dataUrl,
-                     'catId': c.unknownCategoryContainer.categoryIdContainer.categoryId}
-                    for c in data.payload.browseResponse.category]
-
-        output = []
-        clusters = []
-
-        if utils.hasPrefetch(data):
-            for pf in data.preFetch:
-                clusters.extend(pf.response.payload.listResponse.cluster)
-
-        # result contains apps of a specific category
-        # organized by sections
-        for cluster in clusters:
-            for doc in cluster.doc:
-                apps = [a for a in doc.child]
-                apps = list(map(utils.fromDocToDictionary,
-                                apps))
-                section = {'title': doc.title,
-                           'docid': doc.docid,
-                           'apps': apps}
-                output.append(section)
-        return output
+        return utils.parseProtobufObj(data.payload.browseResponse)
 
     def list(self, cat, ctr=None, nb_results=None, offset=None):
-        """List apps for a specfic category *cat*.
+        """List all possible subcategories for a specific category. If
+        also a subcategory is provided, list apps from this category.
 
-        If ctr (subcategory ID) is None, returns a list of valid subcategories.
-
-        If ctr is provided, list apps within this subcategory."""
+        Args:
+            cat (str): category id
+            ctr (str): subcategory id
+            nb_results (int): if a subcategory is specified, limit number
+                of results to this number
+            offset (int): if a subcategory is specified, start counting from this
+                result
+        Returns:
+            A list of categories. If subcategory is specified, a list of apps in this
+            category.
+        """
         path = LIST_URL + "?c=3&cat={}".format(requests.utils.quote(cat))
         if ctr is not None:
             path += "&ctr={}".format(requests.utils.quote(ctr))
         if nb_results is not None:
-            path += "&n={}".format(requests.utils.quote(nb_results))
+            path += "&n={}".format(requests.utils.quote(str(nb_results)))
         if offset is not None:
-            path += "&o={}".format(requests.utils.quote(offset))
+            path += "&o={}".format(requests.utils.quote(str(offset)))
         data = self.executeRequestApi2(path)
         clusters = []
         docs = []
         if ctr is None:
             # list subcategories
             for pf in data.preFetch:
-                clusters.extend(pf.response.payload.listResponse.cluster)
-            for c in clusters:
-                docs.extend(c.doc)
-            return [d.docid for d in docs]
+                for cluster in pf.response.payload.listResponse.doc:
+                    clusters.extend(cluster.child)
+            return [c.docid for c in clusters]
         else:
-            childs = []
-            clusters.extend(data.payload.listResponse.cluster)
-            for c in clusters:
-                docs.extend(c.doc)
-            for d in docs:
-                childs.extend(d.child)
-            return [utils.fromDocToDictionary(c)
-                    for c in childs]
+            apps = []
+            for d in data.payload.listResponse.doc: # categories
+                for c in d.child: # sub-category
+                    for a in c.child: # app
+                        apps.append(utils.parseProtobufObj(a))
+            return apps
 
     def reviews(self, packageName, filterByDevice=False, sort=2,
                 nb_results=None, offset=None):
@@ -518,6 +481,7 @@ class GooglePlayAPI(object):
             dict object containing all the protobuf data returned from
             the api
         """
+        # TODO: select the number of reviews to return
         path = REVIEWS_URL + "?doc={}&sort={}".format(requests.utils.quote(packageName), sort)
         if nb_results is not None:
             path += "&n={}".format(nb_results)
@@ -527,19 +491,8 @@ class GooglePlayAPI(object):
             path += "&dfil=1"
         data = self.executeRequestApi2(path)
         output = []
-        for rev in data.payload.reviewResponse.getResponse.review:
-            author = {'personIdString': rev.author2.personIdString,
-                      'personId': rev.author2.personId,
-                      'name': rev.author2.name,
-                      'profilePicUrl': rev.author2.urls.url,
-                      'googlePlusUrl': rev.author2.googlePlusUrl}
-            review = {'documentVersion': rev.documentVersion,
-                      'timestampMsec': rev.timestampMsec,
-                      'starRating': rev.starRating,
-                      'comment': rev.comment,
-                      'commentId': rev.commentId,
-                      'author': author}
-            output.append(review)
+        for review in data.payload.reviewResponse.getResponse.review:
+            output.append(utils.parseProtobufObj(review))
         return output
 
     def _deliver_data(self, url, cookies):
@@ -640,11 +593,12 @@ class GooglePlayAPI(object):
         """
 
         if self.authSubToken is None:
-            raise Exception("You need to login before executing any request")
+            raise LoginError("You need to login before executing any request")
 
         if versionCode is None:
             # pick up latest version
-            versionCode = self.details(packageName).get('versionCode')
+            appDetails = self.details(packageName).get('details').get('appDetails')
+            versionCode = appDetails.get('versionCode')
 
         headers = self.getHeaders()
         params = {'ot': str(offerType),
@@ -681,6 +635,35 @@ class GooglePlayAPI(object):
         if response.commands.displayErrorMessage != "":
             raise RequestError(response.commands.displayErrorMessage)
 
+    def toc(self):
+        response = requests.get(TOC_URL,
+                               headers=self.getHeaders(),
+                               verify=ssl_verify,
+                               timeout=60,
+                               proxies=self.proxies_config)
+        data = googleplay_pb2.ResponseWrapper.FromString(response.content)
+        tocResponse = data.payload.tocResponse
+        if utils.hasTosContent(tocResponse) and utils.hasTosToken(tocResponse):
+            self.acceptTos(tocResponse.tosToken)
+        if utils.hasCookie(tocResponse):
+            self.dfeCookie = tocResponse.cookie
+        return utils.parseProtobufObj(tocResponse)
+
+
+    def acceptTos(self, tosToken):
+        params = {
+            "tost": tosToken,
+            "toscme": "false"
+        }
+        response = requests.get(ACCEPT_TOS_URL,
+                               headers=self.getHeaders(),
+                               params=params,
+                               verify=ssl_verify,
+                               timeout=60,
+                               proxies=self.proxies_config)
+        data = googleplay_pb2.ResponseWrapper.FromString(response.content)
+        return utils.parseProtobufObj(data.payload.acceptTosResponse)
+
     @staticmethod
     def getDevicesCodenames():
         return config.getDevicesCodenames()
